#!/usr/bin/env python3

"""Blueprints for HTML serving 'ui'."""

import logging
from datetime import timedelta
from importlib.resources import files

from click import group
from flask import (
    Blueprint,
    current_app,
    request,
    Request,
    redirect,
    render_template,
    session,
    url_for,
    flash,
)

from .util import salt, is_logged_in

log = logging.getLogger(__name__)
BLUEPRINT = Blueprint("user", __name__)


@BLUEPRINT.route("/")
def root():
    return (
        render_template(
            "index.html.j2",
        ),
        200,
    )


@BLUEPRINT.route("/user/login", methods=["GET", "POST"])
def login():
    if is_logged_in(request):
        return redirect("/")

    elif request.method == "POST":
        if sid := request.db.try_login(
            username := request.form["username"],
            salt(request.form["password"]),
            timedelta(days=1),
        ):
            resp = redirect("/")
            resp.set_cookie("sid", sid)
            flash(f"Welcome, {username}", category="success")
            return resp

        else:
            flash("Incorrect username/password", category="error")
            return render_template("login.html.j2")

    else:
        return render_template("login.html.j2")


@BLUEPRINT.route("/user/register", methods=["GET", "POST"])
def register():
    if is_logged_in(request):
        return redirect("/")

    elif request.method == "POST":
        try:
            username = request.form["username"]
            email = request.form["email"]
            group_id = None
            status_id = None

            for user_config in current_app.config.get("users", []):
                if user_config["email"] == email:
                    if "group_id" in user_config:
                        group_id = user_config["group_id"]

                    if "status_id" in user_config:
                        status_id = user_config["status_id"]

                    break

            if res := request.db.try_create_user(
                username, email, salt(request.form["password"]), group_id, status_id
            ):
                id, status = res
                if status == -1:
                    flash(
                        "Please check your email for a verification request",
                        category="success",
                    )
                return render_template("register.html.j2")

        except Exception as e:
            log.exception("Error encountered while registering a user...")

        flash("Unable to register that username", category="error")
        return render_template("register.html.j2")

    else:
        return render_template("register.html.j2")


@BLUEPRINT.route("/user/logout")
def logout():
    # Invalidate the user's authorization
    request.db.delete_key(request.sid)
    resp = redirect("/")
    resp.set_cookie("sid", "")
    return resp


@BLUEPRINT.route("/user", methods=["GET", "POST"])
def settings():
    if is_logged_in(request):
        return redirect("/")

    elif request.method == "POST":
        pass

    else:
        return render_template("user.html.j2")
